import polars as pl
from polars_textproc import repetition_signals, minhash, scrub, uuid4, samplebyte, compression_ratio, register_namespace

from tokenizers import Tokenizer

tokenizer = Tokenizer.from_pretrained('bert-base-cased')

register_namespace()

HEAD = 10_000
english = """\
English is a West Germanic language in the Indo-European language family, whose speakers, called Anglophones, originated in early medieval England on the island of Great Britain.[4][5][6] The namesake of the language is the Angles, one of the Germanic peoples that migrated to Britain after its Roman occupiers left. English is the most spoken language in the world, primarily due to the global influences of the former British Empire (succeeded by the Commonwealth of Nations) and the United States.[7] English is the third-most spoken native language, after Mandarin Chinese and Spanish;[8] it is also the most widely learned second language in the world, with more second-language speakers than native speakers.

English is either the official language or one of the official languages in 59 sovereign states (such as India, Ireland, and Canada). In some other countries, it is the sole or dominant language for historical reasons without being explicitly defined by law (such as in the United States and United Kingdom).[9] It is a co-official language of the United Nations, the European Union, and many other international and regional organisations. It has also become the de facto lingua franca of diplomacy, science, technology, international trade, logistics, tourism, aviation, entertainment, and the Internet.[10] English accounts for at least 70% of total native speakers of the Germanic languages, and Ethnologue estimated that there were over 1.5 billion speakers worldwide as of 2021.[3]"""

swedish = """\
Engelska (English) är ett västgermanskt språk, dock starkt påverkat av bland annat franska och latin. Det är världens mest spridda språk och fungerar i många delar av världen som lingua franca.[2]
Historia

Engelskan är ett västgermanskt språk. Det nu talade västgermanska språk som räknas som närmast besläktat med engelskan är lågskotskan (Scots) och därefter frisiskan.[3] Uppemot 7 500 ord i engelskans aktiva ordförråd är franska lånord som en följd av den normandiska invasionen av England år 1066 efter slaget vid Hastings.[4] Germanska språkvarieteter kom även tidigare till de brittiska öarna på 400-talet med olika germanska stammar. Dessa har traditionellt beskrivits som främst saxare (talande saxiska dialekter) och angler (talande angliska dialekter) som tillsammans fått beteckningen anglosaxare, samt även jutar, friser och franker.[4] Britannien hade tidigare varit bebott främst av kelter, som talade keltiska språk, men dessa trängdes undan av germanerna. De keltiska folken fortsatte länge att vara dominerande i Skottland och Wales samt på Irland. De keltiska språken återfinns även i modern tid som iriska, skotsk gaeliska och kymriska, och fanns t.o.m. år 1777 i Cornwall."""

interlingua = """\
Le lingua anglese[1] es un lingua germanic con influentia del lingua latin, gratias al influentia del lingua francese diffundite durante le regno normanne del Anglaterra a partir de 1066.
Stato

Iste lingua non es parlate per le major numero del humanos in le mundo (vide lingua chinese mandarin), ma illo es inseniate in tote le mundo. Illo ha quasi devenite un lingua mundial, sovente usate como lingua franca. Il ha de 347.600.000 a 580.000.000 de parlantes in le mundo del quales circa 60.000.000 vive in Europa. Le anglese es un del linguas fontes primari de Interlingua e un del linguas official del Union Europee e del Nationes Unite. """

interlingua2 = """\
Le lingua anglese[1] es un lingua germanic con influentia del lingua latin, gratias al influentia del lingua francese diffundite durante le regno normanne del Anglaterra a partir de 1066.
Stato

Iste lingua non es parlate per le major numero del humanos in le mundo (vide lingua chinese mandarin), ma illo es inseniate in tote le mundo. Illo ha quasi devenite un lingua mundial, sovente usate como lingua franca. Il ha de 347.600.000 a 580.000.000 de parlantes in le mundo del quales circa 60.000.000 vive in Europa. Le anglese es un del linguas fontes primari de Interlingua e un del linguas official del Union Europee e del Nationes Unite. 
And some extra stuff."""
repetetive = """\
This is a very repetetive text that is very repetetive and a text
""" * 10

empty = ""

single = "a"

df = pl.DataFrame({
    "text": [english, swedish, interlingua, interlingua2, None, repetetive, empty, single], 
    "num" : [1, 2, 3, 4, 5, 6, 7, 8],
    })

bsize_str = 2 * 128//8

def minhash_buckets(col, buckets, bsize):
    hashes = minhash(col, hashes=buckets*bsize)
    return [hashes.str.slice(i*bsize_str, bsize_str).alias(f'bucket_{i}') for i in range(buckets)]

normalized = pl.col('text').textproc.normalize(form='NFC', lowercase=True, only_script=True, remove_newlines=True, contract_whitespace=True)
buckets = 14
bsize = 8
hashes = buckets * bsize
lf = df.lazy()

text = pl.col('text')

lf = lf.with_columns(
    id = uuid4(pl.first()),
    samplebyte = samplebyte(pl.first()),
    norm = normalized, 
    repetition=repetition_signals(normalized, tokenizer_pattern=r'.'), 
    langid=text.textproc.fasttext(path="model.bin", labels=["__label__swe_Latn", "__label__eng_Latn"]), 
    minhash=minhash(normalized, buckets=buckets, bsize=bsize),
    redacted=scrub('text', patterns=[r'\bE\w*\b']),
    compression_ratio=compression_ratio('text'),
    compressed_size=text.textproc.compressed_size(),
    tokenized=text.textproc.tokenize(tokenizer=tokenizer),
)
lf = lf.with_columns([pl.col('minhash').str.slice(i*bsize_str, bsize_str).alias(f'bucket_{i}') for i in range(buckets)])
df = lf.collect()
print(df)
print(df[0, 'bucket_0'])
print(df[0, 'bucket_13'])
print(df[0, 'minhash'])
print(df[:, 'redacted'])
print(df.select('compression_ratio', 'compressed_size', pl.col('text').str.len_bytes()))
print(df.select('repetition').unnest('repetition'))
print(df.select('tokenized'))
print(df.select('norm')[0, 0])



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List

from tokenizers import Tokenizer

import polars as pl
from polars.plugins import register_plugin_function

from polars_textproc._internal import __version__ as __version__

if TYPE_CHECKING:
    from polars_textproc.typing import IntoExprColumn

LIB = Path(__file__).parent

SEED = [
    179,
    28,
    18,
    84,
    75,
    144,
    79,
    252,
    138,
    70,
    118,
    68,
    23,
    234,
    55,
    243,
    220,
    195,
    42,
    178,
    73,
    192,
    91,
    161,
    228,
    176,
    67,
    210,
    33,
    75,
    126,
    56,
]


def tokenize(expr: IntoExprColumn, *, tokenizer: Tokenizer | str) -> pl.Expr:
    if isinstance(tokenizer, Tokenizer):
        kwargs = {"payload": tokenizer.to_str(), "is_path": False}
    elif isinstance(tokenizer, str):
        kwargs = {
            "payload": tokenizer,
            "is_path": True,
        }
    else:
        raise ValueError(
            f"tokenizer must be str or Tokenizer instance, found: {type(tokenizer)}"
        )
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="tokenize",
        is_elementwise=True,
        kwargs=kwargs,
    )


def compressed_size(expr: IntoExprColumn, *, level: int = 6) -> pl.Expr:
    assert 0 <= level <= 9, "compression level must be between 0 and 9, not "
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="compressed_size",
        is_elementwise=True,
        kwargs={
            "level": level,
        },
    )


def compression_ratio(expr: IntoExprColumn, *, level: int = 6) -> pl.Expr:
    assert 0 <= level <= 9, "compression level must be between 0 and 9, not "
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="compression_ratio",
        is_elementwise=True,
        kwargs={
            "level": level,
        },
    )


def samplebyte(expr: IntoExprColumn) -> pl.Expr:
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="samplebyte",
        is_elementwise=True,
    )


def uuid4(expr: IntoExprColumn) -> pl.Expr:
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="uuid4",
        is_elementwise=True,
    )


def minhash(
    expr: IntoExprColumn,
    *,
    tokenizer_pattern: str = r"\w+",
    seed=SEED,
    buckets=14,
    bsize=8,
    window=5,
) -> pl.Expr:
    """
    construct a hex representation of the minhash hash of the given text column.
    `tokenizer_pattern`: tokenizer pattern for the word-shingling.
    `seed`: The seed for the hash-permutations.
    `buckets`: Number of minhash buckets.
    `bsize`: Size (in hashes) of each minhash bucket.
    `window`: Shingle window size.

    By default, it creates `buckets * bsize` 64-bit hashes, and then hash each bucket into
    a 128-bit hash, which is then hex encoded as a string, resulting in a `buckets*32` byte
    long minhash signature (With every chunk of 32 bytes being a separate bucket signature).

    Alternatively, if bsize is set to one, it skips the 128-bit hashing, resulting in a
    `buckets*16` byte long minhash signature.

    Suppling buckets > 1 saves space in the case where you've commited to a specific bucketing
    strategy, and the 128-bit hash should be enough to avoid collisions.

    Supplying buckets == 1 could be useful for varying bucket sizes.
    """
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="minhash",
        is_elementwise=True,
        kwargs={
            "tokenizer_pattern": tokenizer_pattern,
            "buckets": buckets,
            "bsize": bsize,
            "seed": seed,
            "window": window,
        },
    )


def repetition_signals(
    expr: IntoExprColumn, *, tokenizer_pattern: str = r"\w+", num_top=4, num_dup=10
) -> pl.Expr:
    """
    Runs gopher repetition signals on the given text column.
    Words are extracted using the supplied tokenizer pattern.
    Computes "top_n_gram_char_ratio"-signals for top_1, top_2, .. top_{num_top}.
    Computes "dup_n_gram_char_ratio"-signals for dup_{num_top+1}, .. dup_{num_dup}.
    """
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="repetition_signals",
        is_elementwise=True,
        kwargs={
            "tokenizer_pattern": tokenizer_pattern,
            "num_top": num_top,
            "num_dup": num_dup,
        },
    )


def scrub(
    expr: IntoExprColumn, *, patterns: List[str], replacement: str = "REDACTED"
) -> pl.Expr:
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="scrub",
        is_elementwise=True,
        kwargs={"patterns": patterns, "replacement": replacement},
    )


def fasttext(
    expr: IntoExprColumn,
    *,
    path: str,
    labels: List[str],
    output_aggregate: bool = True,
    output_scores: bool = False,
) -> pl.Expr:
    """
    Runs a fasttext model against the given text column.
    `path` is the path to the fasttext model bin path.
    `labels` are the labels that should be included in the output.
    output_aggregate=True =>
        output columns:
        `top_label`   : String = top scoring label
        `top_score`   : Float  = score of the top scoring label
        `total_score` : Float  = total score of all given labels
    output_scores=True =>
        output columns:
        `$label` : Float = score of the `$label` (only including given `labels`)
    """
    return register_plugin_function(
        args=[expr],
        plugin_path=LIB,
        function_name="fasttext",
        is_elementwise=True,
        kwargs={
            "path": path,
            "labels": labels,
            "output_aggregate": output_aggregate,
            "output_scores": output_scores,
        },
    )


def register_namespace(name="textproc"):
    @pl.api.register_expr_namespace(name)
    class TextprocNamespace:
        def __init__(self, expr: pl.Expr) -> None:
            self._expr = expr

        def normalize(
            self,
            *,
            form=None,
            lowercase=False,
            only_script=False,
            remove_newlines=False,
            contract_whitespace=False,
        ) -> pl.Expr:
            expr = self._expr
            if form:
                expr = expr.str.normalize(form=form)
            if lowercase:
                expr = expr.str.to_lowercase()
            if only_script:
                expr = expr.str.replace_all(r"[\W&&\S]+", " ")
            if remove_newlines:
                expr = expr.str.replace_all(r"\n", " ")
            if contract_whitespace:
                expr = expr.str.replace_all(r"[\s--\n]+", " ")
            return expr

        def compressed_size(self, *, level: int = 6) -> pl.Expr:
            return compressed_size(self._expr, level=level)

        def compression_ratio(self, *, level: int = 6) -> pl.Expr:
            return compression_ratio(self._expr, level=level)

        def samplebyte(self) -> pl.Expr:
            return samplebyte(self._expr)

        def uuid4(self) -> pl.Expr:
            return uuid4(self._expr)

        def tokenize(self, *, tokenizer: Tokenizer | str) -> pl.Expr:
            return tokenize(self._expr, tokenizer=tokenizer)

        def minhash(
            self,
            *,
            tokenizer_pattern: str = r"\w+",
            seed=SEED,
            buckets=14,
            bsize=8,
            window=5,
        ) -> pl.Expr:
            return minhash(
                self._expr,
                tokenizer_pattern=tokenizer_pattern,
                seed=seed,
                buckets=buckets,
                bsize=bsize,
                window=window,
            )

        def repetition_signals(
            self,
            *,
            tokenizer_pattern: str = r"\w+",
            num_top=4,
            num_dup=10,
        ) -> pl.Expr:
            return repetition_signals(
                self._expr,
                tokenizer_pattern=tokenizer_pattern,
                num_top=num_top,
                num_dup=num_dup,
            )

        def scrub(
            self,
            *,
            patterns: List[str],
            replacement: str = "REDACTED",
        ) -> pl.Expr:
            return scrub(self._expr, patterns=patterns, replacement=replacement)

        def fasttext(
            self,
            *,
            path: str,
            labels: List[str],
            output_aggregate: bool = True,
            output_scores: bool = False,
        ) -> pl.Expr:
            return fasttext(
                self._expr,
                path=path,
                labels=labels,
                output_aggregate=output_aggregate,
                output_scores=output_scores,
            )
